# Ngaro VM
# Copyright (c) 2010, Charles Childers
# -----------------------------------------------------------------------------
@ip = 0                # instruction pointer
@sp = 0                # stack pointer
@rp = 0                # return pointer
@stack   = {}          # data stack
@address = {}          # return stack
@memory  = {}          # simulated ram
@ports   = {}          # io ports
@files   = {}          # file handles
# -----------------------------------------------------------------------------
def handle_devices()
  if @ports[0] == 1 then return end
  if @ports[0] == 0 then
    @ports[0] = 1
    if @ports[1] == 1 then
      @ports[1] = STDIN.getc
      if @ports[1] == 13 then @ports[1] = 10 end
    end
    if @ports[2] == 1 then
      if @stack[@sp] < 0 then
        puts "\e[H\e[2J"
      elsif @stack[@sp] > 0 and @stack[@sp] < 128 then
        print @stack[@sp].chr
        STDOUT.flush
      end
      @sp = @sp - 1
      @ports[2] = 0
    end

    if @ports[4] == -1 then   # file open
      slot = 0
      i = 1
      while i < 8 do
        if @files[i] == 0 then
          slot = i
        end
        i = i + 1
      end
      mode = @stack[@sp]
      @sp = @sp - 1
      name = ""
      i = @stack[@sp]
      @sp = @sp - 1
      while @memory[i] != 0 do
        name = name + @memory[i].chr
        i = i + 1
      end
      if mode == 0 then
        @files[slot] = File.open(name, 'r')
      end
      if mode == 1 then
        @files[slot] = File.open(name, 'w')
      end
      if mode == 2 then
        @files[slot] = File.open(name, 'a')
      end
      if mode == 3 then
        @files[slot] = File.open(name, 'r+')
      end
      @ports[4] = slot
    end
    if @ports[4] == -2 then   # file read
      slot = @stack[@sp]
      @sp = @sp - 1
      @ports[4] = @files[slot].readchar
    end
    if @ports[4] == -3 then   # file write
      slot = @stack[@sp]
      @sp = @sp - 1
      key = @stack[@sp]
      @sp = @sp - 1
      @files[slot].write(key.chr)
      @ports[4] = 1
    end
    if @ports[4] == -4 then   # file close
      slot = @stack[@sp]
      @sp = @sp - 1
      @files[slot].close
      @files[slot] = 0
      @ports[4] = 0
    end
    if @ports[4] == -5 then   # file position
      slot = @stack[@sp]
      @sp = @sp - 1
      @ports[4] = @files[slot].pos
    end
    if @ports[4] == -6 then   # file seek
      slot = @stack[@sp]
      @sp = @sp - 1
      offset = @stack[@sp]
      @sp = @sp - 1
      @files[slot].seek(offset, IO::SEEK_SET)
      @ports[4] = 0
    end
    if @ports[4] == -7 then   # file size
      slot = @stack[@sp]
      @sp = @sp - 1
      offset = @files[slot].pos
      @files[slot].seek(0, IO::SEEK_END)
      @ports[4] = @files[slot].pos
      @files[slot].seek(offset, IO::SEEK_SET)
    end
    if @ports[4] == -8 then   # file delete
      name = ""
      i = @stack[@sp]
      @sp = @sp - 1
      while @memory[i] != 0 do
        name = name + @memory[i].chr
        i = i + 1
      end
      File.delete(name)
      @ports[4] = 0
    end

    if @ports[5] == -1 then  # memory size
      @ports[5] = 1000000
    end
    if @ports[5] == -2 then  # canvas exists?
      @ports[5] = 0
    end
    if @ports[5] == -3 then  # canvas width
      @ports[5] = 0
    end
    if @ports[5] == -4 then  # canvas height
      @ports[5] = 0
    end
    if @ports[5] == -5 then  # stack depth
      @ports[5] = @sp
    end
    if @ports[5] == -6 then  # address stack depth
      @ports[5] = @rp
    end
    if @ports[5] == -7 then  # mouse exists?
      @ports[5] = 0
    end
    if @ports[5] == -8 then  # time
      @ports[5] = Time.now.to_i
    end
    if @ports[5] == -9 then  # exit vm
      @ip = 1000000
      @ports[5] = 0
    end
    if @ports[5] == -10 then # environment
      @ports[5] = 0
    end
    if @ports[5] == -11 then # console width
      @ports[5] = 0
    end
    if @ports[5] == -12 then # console height
      @ports[5] = 0
    end
  end
end
# -----------------------------------------------------------------------------
def process()
  opcode = @memory[@ip]
  if    opcode ==  0 then  # nop
  elsif opcode ==  1 then  # lit
    @sp = @sp + 1
    @ip = @ip + 1
    @stack[@sp] = @memory[@ip]
  elsif opcode ==  2 then  # dup
    @sp = @sp + 1
    @stack[@sp] = @stack[@sp - 1]
  elsif opcode ==  3 then  # drop
    @sp = @sp - 1
  elsif opcode ==  4 then  # swap
    a = @stack[@sp]
    @stack[@sp] = @stack[@sp - 1]
    @stack[@sp - 1] = a
  elsif opcode ==  5 then  # push
    @rp = @rp + 1
    @address[@rp] = @stack[@sp]
    @sp = @sp - 1
  elsif opcode ==  6 then  # pop
    @sp = @sp + 1
    @stack[@sp] = @address[@rp]
    @rp = @rp - 1
  elsif opcode ==  7 then  # loop
    @stack[@sp] = @stack[@sp] - 1
    if @stack[@sp] != 0 and @stack[@sp] > -1 then
      @ip = @ip + 1
      @ip = @memory[@ip] - 1
    else
      @ip = @ip + 1
      @sp = @sp - 1
    end
  elsif opcode ==  8 then  # jump
    @ip = @ip + 1
    @ip = @memory[@ip] - 1
    if @memory[@ip + 1] == 0 then @ip = @ip + 1 end
    if @memory[@ip + 1] == 0 then @ip = @ip + 1 end
  elsif opcode ==  9 then  # return
    @ip = @address[@rp]
    @rp = @rp - 1
  elsif opcode == 10 then  # >= jump
    @ip = @ip + 1
    if @stack[@sp - 1] > @stack[@sp] then
      @ip = @memory[@ip] - 1
    end
    @sp = @sp - 2
  elsif opcode == 11 then  # <= jump
    @ip = @ip + 1
    if @stack[@sp - 1] < @stack[@sp] then
      @ip = @memory[@ip] - 1
    end
    @sp = @sp - 2
  elsif opcode == 12 then  # != jump
    @ip = @ip + 1
    if @stack[@sp - 1] != @stack[@sp] then
      @ip = @memory[@ip] - 1
    end
    @sp = @sp - 2
  elsif opcode == 13 then  # == jump
    @ip = @ip + 1
    if @stack[@sp - 1] == @stack[@sp] then
      @ip = @memory[@ip] - 1
    end
    @sp = @sp - 2
  elsif opcode == 14 then  # @
    @stack[@sp] = @memory[@stack[@sp]]
  elsif opcode == 15 then  # !
    @memory[@stack[@sp]] = @stack[@sp - 1]
    @sp = @sp - 2
  elsif opcode == 16 then  # +
    @stack[@sp - 1] = @stack[@sp - 1] + @stack[@sp]
    @sp = @sp - 1
  elsif opcode == 17 then  # -
    @stack[@sp - 1] = @stack[@sp - 1] - @stack[@sp]
    @sp = @sp - 1
  elsif opcode == 18 then  # *
    @stack[@sp - 1] = @stack[@sp - 1] * @stack[@sp]
    @sp = @sp - 1
  elsif opcode == 19 then  # /mod
    a = @stack[@sp]
    b = @stack[@sp - 1]
    @stack[@sp] = (b / a).floor
    @stack[@sp - 1] = b % a
  elsif opcode == 20 then  # and
    a = @stack[@sp]
    b = @stack[@sp - 1]
    @sp = @sp - 1
    @stack[@sp] = b & a
  elsif opcode == 21 then  # or
    a = @stack[@sp]
    b = @stack[@sp - 1]
    @sp = @sp - 1
    @stack[@sp] = b | a
  elsif opcode == 22 then  # xor
    a = @stack[@sp]
    b = @stack[@sp - 1]
    @sp = @sp - 1
    @stack[@sp] = b ^ a
  elsif opcode == 23 then  # <<
    a = @stack[@sp]
    b = @stack[@sp - 1]
    @sp = @sp - 1
    @stack[@sp] = b << a
  elsif opcode == 24 then  # >>
    a = @stack[@sp]
    b = @stack[@sp - 1]
    @sp = @sp - 1
    @stack[@sp] = b >> a
  elsif opcode == 25 then  # 0;
    if @stack[@sp] == 0 then
      @sp = @sp - 1
      @ip = @address[@rp]
      @rp = @rp - 1
    end
  elsif opcode == 26 then  # inc
    @stack[@sp] = @stack[@sp] + 1
  elsif opcode == 27 then  # dec
    @stack[@sp] = @stack[@sp] - 1
  elsif opcode == 28 then  # in
    @stack[@sp] = @ports[@stack[@sp]]
  elsif opcode == 29 then  # out
    @ports[@stack[@sp]] = @stack[@sp - 1]
    @sp = @sp - 2
  elsif opcode == 30 then  # wait
    handle_devices()
  else                     # call (implicit)
    @rp = @rp + 1
    @address[@rp] = @ip
    @ip = @memory[@ip] - 1
    if @memory[@ip + 1] == 0 then @ip = @ip + 1 end
    if @memory[@ip + 1] == 0 then @ip = @ip + 1 end
  end
end
# -----------------------------------------------------------------------------
@cells = (File.size("retroImage") / 4)
File.open("retroImage", "r") do |f|
  (0..(@cells-1)).each do |i|
    @memory[i] = f.sysread(4).unpack("i")[0]
  end
end
# -----------------------------------------------------------------------------
(@cells..1000000).each do |i|
  @memory[i] = 0
end
(0..1024).each do |i|
  @stack[i] = 0
  @address[i] = 0
  @ports[i] = 0
end
# -----------------------------------------------------------------------------
while @ip < 1000000 do
  process()
  @ip = @ip + 1
end

